#include "jis208.hh"

#include "unicode/jis0201.h"
#include "unicode/jis0208.h"
#include "unicode/jis0212.h"

jis208::jis208() : instate(ascii), outstate(ascii)
{
	jis201in = new jis201_to_uni;
	jis201out= new uni_to_jis201;
	jis208in = new jis208_to_uni;
	jis208out= new uni_to_jis208;
	jis212in = new jis212_to_uni;
	jis212out= new uni_to_jis212;
}

jis208::~jis208()
{
	delete jis201in; delete jis201out;
	delete jis208in; delete jis208out;
	delete jis212in; delete jis212out;
}

void jis208::setoutstate(string &ss, jistate newstate)
{
    if(outstate != newstate)
    {
        outstate = newstate;
        switch(outstate)
        { 
            case ascii:      ss += "\33(B"; break;
            case st201latin: ss += "\33(J"; break;
            case st201kana:  ss += "\33(I"; break;
            case st78:       ss += "\33$@"; break;
            case st83:       ss += "\33$B"; break;
            case st212:      ss += "\33$(D"; break;
            default: ;
        }
    }
}

const wstring jis208::c2uni(const string &s, bool html)
{
    wstring ss;
    unsigned a, b=s.size();
    charmap::const_iterator i;
    
    for(a=0; a<b; ++a)
    {
        if(s[a]==jis208_SO) { instate=st201kana; }
        else if(s[a]==jis208_SI) { if(instate != ascii && instate != st201latin) instate=ascii; }
        else if(s[a]=='\33') { instate=gotesc; continue; }
        
        switch(instate)
        {
            case gotesc:
                if(s[a]=='(') { instate = gotsulku; break; }
                if(s[a]=='$') { instate = gotdolla; break; }
                instate = ascii; 
                /* fallthru */
            case ascii:
                if(html)
                {
                    ss += TryParseEntities(s, a);
                    if(a==b)break;
                }
                if((unsigned char)s[a] < 0x80)
                {
                    ss += (unsigned char)s[a];
                    break;
                }
                /* fallthru */
            case st201latin:
                i = jis201in->find((unsigned char)s[a]);
                if(i == jis201in->end())
                    ss += 0xFFFD; /* undefined char */
                else
                    ss += i->second;
                break;
            case st201kana:
                i = jis201in->find((unsigned char)s[a] | 0x80);
                if(i == jis201in->end())
                    ss += 0xFFFD; /* undefined char */
                else
                    ss += i->second;
                break;
            case gotsulku:
                instate = ascii;
                if(s[a]=='J') instate = st201latin;
                else if(s[a]=='I') instate = st201kana;
                break;
            case gotsulku2:
                instate = ascii;
                if(s[a]=='D') instate = st212;
                break;
            case gotdolla:
                instate = ascii;
                if(s[a]=='@' || s[a]=='B') instate = st78a83;
                else if(s[a]=='(') instate = gotsulku2;
                break;
            case st78a83:
                cache = s[a] & 0x7F;
                instate = st78b83;
                break;
            case st78b83:
            {
                wchar_t t = cache * 256 + ((unsigned char)s[a] & 0x7F);
                i = jis208in->find(t);
                if(i == jis208in->end())
                    ss += 0xFFFD; /* undefined char */
                else
                    ss += i->second;
                instate = st78a83;
                break;
            }
            case st212a:
                cache = s[a] & 0x7F;
                instate = st212b;
                break;
            case st212b:
            {
                wchar_t t = cache * 256 + ((unsigned char)s[a] & 0x7F);
                i = jis212in->find(t);
                if(i == jis212in->end())
                    ss += 0xFFFD; /* undefined char */
                else
                    ss += i->second;
                instate = st212a;
                break;
            }
        }
    }
    return ss;
}

const string jis208::uni2c(const wstring &s)
{
    string ss;
    unsigned a, b=s.size();
    for(a=0; a<b; ++a)
    {
        charmap::const_iterator i;
        
        if(s[a] < 0x20 || s[a] > 0x7D)
        {
            /* jis201? */
            i = jis201out->find(s[a]);
            if(i != jis201out->end())
            {
                if(i->second < 0x80)
                {
                    setoutstate(ss, st201latin);
                    ss += (char)i->second;
                }
                else
                {
                    setoutstate(ss, st201kana);
                    ss += (char)(i->second & 127);
                }
                continue;
            }
        }
        
        /* latin1?  */
        if(s[a] < 0x100)
        {
            setoutstate(ss, ascii);
            ss += (char)s[a];
            continue;
        }
        
        /* jis208? */
        i = jis208out->find(s[a]);
        if(i != jis208out->end())
        {
            setoutstate(ss, st83);
            ss += (char)(i->second >> 8);
            ss += (char)(i->second & 127);
            continue;
        }
        
        /* jis212? */
        i = jis212out->find(s[a]);
        if(i != jis212out->end())
        {
            setoutstate(ss, st212);
            ss += (char)(i->second >> 8);
            ss += (char)(i->second & 127);
            continue;
        }
        
        setoutstate(ss, ascii);
        ss += '?';
    }
    
    /* Always end with ascii mode */
    setoutstate(ss, ascii);
    return ss;
}
