#define bqt_unicode_cc

#include <cstdio>
#include "charsets.hh"

static bool similar_char(char c, char c2)
{
    if(c==c2)return true;
    if((c =='-' || c=='_')
     &&(c2=='-' ||c2=='_'))return true;
    return false;
}

/* Adapted from PHP */
static void similar_str(const char *txt1, unsigned len1,
                        const char *txt2, unsigned len2,
                        unsigned &pos1, unsigned &pos2, unsigned &max)
{
    const char *p, *q;
    const char *end1 = txt1 + len1;
    const char *end2 = txt2 + len2;
    unsigned l;
    
    max = 0;
    for (p = txt1; p < end1; p++)
    {
        for (q = txt2; q < end2; q++)
        {
            for (l = 0; (p + l < end1) && (q + l < end2) && similar_char(p[l], q[l]); 
                 l++);
            if (l > max)
            {
                max = l;
                pos1 = p - txt1;
                pos2 = q - txt2;
            }
        }
    }
}

/* Adapted from PHP */
static int similar_char(const char *txt1, unsigned len1, 
                        const char *txt2, unsigned len2)
{
    unsigned pos1, pos2, max;

    similar_str(txt1, len1, txt2, len2, pos1, pos2, max);
    
    unsigned sum = max;    
    if(sum)
    {
        if(pos1 && pos2)
            sum += similar_char(txt1, pos1, txt2, pos2);
        if((pos1 + max < len1) && (pos2 + max < len2))
            sum += similar_char(txt1 + pos1 + max, len1 - pos1 - max, 
                                txt2 + pos2 + max, len2 - pos2 - max);
    }
    return sum;
}

static int csetcompare(const string &csetname, const string &key)
{
    int sim = similar_char(csetname.c_str(), csetname.size(),
                           key.c_str(), key.size());
    return sim * 1000000 / (csetname.size() + key.size());
}

#define INCLUDECONV
#define INCLUDEDATA

#include "unicode/converters.hh"

#undef INCLUDECONV
#undef INCLUDEDATA

charset *newconverter(const string &csetname)
{
	/* Start constants */
    static const char *const csets[] =
    {
#define CONVLIST

#include "unicode/converters.hh"

#undef CONVLIST
    };

    int ind = 0;
    if(!csetname.size())
    {
PickDefault:
        return new easycharset(new iso01_to_uni, new uni_to_iso01);
    }
    
#define CSETKEYCOUNT (sizeof(csets) / sizeof(csets[0]))
    int most=0; int mostind=0;
    for(unsigned a=0; a<CSETKEYCOUNT; ++a)
    {
        if(csetname == csets[a])goto Found;
        
        int score = csetcompare(csetname, csets[a]);
        
        if(score > most)
        {
            most = score;
            mostind = a;
        }
        ++ind;
    }
    fprintf(stderr, "Charconv: Warning: Assuming '%s' means '%s'\n",
        csetname.c_str(), csets[mostind]);
    ind = mostind;

Found:

    switch(ind)
    {

#define CASELIST

#include "unicode/converters.hh"

#undef CASELIST

        default:
            goto PickDefault;
    }
    
    /* Unreachable */
    return NULL;
}
