#include "utf7.hh"

/* Returns true if the next character must be
 * base64-encoded
 */ 
const bool utf7::mustencode(unsigned c) const
{
    if(c=='\n' || c=='\r' || c=='\t' || c==' ')return false;
    
    if(!modified && (c=='=' || c=='\\' || c==0x7E))return true;
    
    return c<0x20 || c>0x7E;
}

const char utf7::base64(unsigned n) const /* n=0..63 */
{
    return (n<26) ? 'A'+n
         : (n<52) ? 'a'+n-26
         : (n<62) ? '0'+n-52
         : (n==62) ? '+'
         : (modified ? ',' : '/');
}

const unsigned utf7::esab64(char n) const /* return=0..64 (64=invalid) */
{
    return (n>='A' && n<='Z') ? n-'A'
         : (n>='a' && n<='z') ? n-'a'+26
         : (n>='0' && n<='9') ? n-'0'+52
         : (n=='+') ? 62
         : (modified && n==',') ? 63
         : (!modified && n=='/') ? 63
         : 64;
}

const wstring utf7::c2uni(const string &s, bool html)
{
    wstring ss;
    unsigned a, b=s.size();
    for(a=0; a<b; ++a)
    {
        unsigned char c = s[a];
        if(!inside)
        {
            if(html)
            {
                ss += TryParseEntities(s, a);
                if(a==b)break;
            }
            c = s[a];
            
            if(c == startchar())
            {
                inside = true;
                cache = 0;
            }
            else
            {
                had = false;
                cachelen = 0;
                cache = c;
            }
        }
        else
        {
            unsigned val = esab64(c);
            if(val == 64) /* not base64? */
            {
                inside = false;
                if(c != endchar()) /* Not terminating char? */
                {
                    ss += c; /* Inline it */
                }
                if(had)
                {
                    /* The cache garbage must be ignored */
                    continue;
                }
                /* was an escaped startchar */
                cache = startchar();               
            }
            else
            {
                had = true;
                cache = cache*64 + val;
                cachelen += 6;
                if(cachelen >= 16)
                {
                    ss += (wchar_t) (cache >> (cachelen -= 16));
                    cache &= ((1 << cachelen)-1);
                }
            }
        }
        if(!inside)
        {
            ss += (wchar_t)cache;
        }
    }
    return ss;
}

const string utf7::uni2c(const wstring &s)
{
    string ss;
    for(unsigned a=0; a<s.size(); ++a)
    {
        if(s[a] == startchar())
        {
            ss += startchar();
            ss += endchar();
            continue;
        }
        if(!mustencode(s[a]))
        {
            ss += (unsigned char)s[a];
            continue;
        }
        ss += startchar();
        unsigned b = 0, bp = 0;
        while(a < s.size())
        {
            if(!mustencode(s[a]))
            {
                /* Cheating... don't switch to ascii mode for one char */
                if(a+1 >= s.size()
                || !mustencode(s[a+1])
                || esab64(s[a]) >= 64
             // || s[a]  == startchar()
                  )
                    break;
            }
            unsigned ch = s[a++];
            bp += 16;
            b <<= 16;
            b |= ch;
            while(bp >= 6)
            {
                unsigned tok = b >> (bp -= 6);
                ss += base64(tok);
                b &= ((1 << bp)-1);
            }
        }
        if(bp > 0)
        {
            /* Trailing garbage */
            ss += base64(b << (6-bp));
        }            
        if(a >= s.size())
            ss += endchar();
        else
        {
            if(esab64(s[a]) < 64)
                ss += endchar();
            ss += (unsigned char)s[a];
        }
    }
    return ss;
}
